/**
 * Convert a glTF to binary glTF.
 *
 * The glTF is expected to have a single buffer and all embedded resources stored in bufferViews.
 *
 * @param {Object} gltf The glTF asset.
 * @param {Uint8Array} binaryBuffer The binary buffer.
 * @returns {Uint8Array} The glb buffer.
 *
 * @private
 */
export function gltfToGlb(gltf, binaryBuffer) {
  const buffer = gltf.buffers[0];
  if (typeof buffer.uri !== "undefined") {
    binaryBuffer = new Uint8Array(0);
  }

  // Create padded binary scene string

  let jsonString = JSON.stringify(gltf);
  const remainder = jsonString.length % 4;
  const padding = remainder === 0 ? 0 : 4 - remainder;
  let whitespace = "";
  for (let i = 0; i < padding; ++i) {
    whitespace += " ";
  }
  jsonString += whitespace;

  const jsonBuffer = new TextEncoder().encode(jsonString);

  // Allocate buffer (Global header) + (JSON chunk header) + (JSON chunk) + (Binary chunk header) + (Binary chunk)
  const glbLength = 12 + 8 + jsonBuffer.length + 8 + binaryBuffer.length;
  const glb = new Uint8Array(glbLength);
  const dataView = new DataView(glb.buffer);

  // Write binary glTF header (magic, version, length)
  let byteOffset = 0;
  dataView.setUint32(byteOffset, 0x46_54_6c_67, true);
  byteOffset += 4;
  dataView.setUint32(byteOffset, 2, true);
  byteOffset += 4;
  dataView.setUint32(byteOffset, glbLength, true);
  byteOffset += 4;
  dataView.setUint32(byteOffset, jsonBuffer.length, true);
  byteOffset += 4;
  dataView.setUint32(byteOffset, 0x4e_4f_53_4a, true);
  byteOffset += 4;

  for (const value of jsonBuffer) {
    dataView.setUint8(byteOffset, value);
    byteOffset += 1;
  }

  dataView.setUint32(byteOffset, binaryBuffer.length, true);
  byteOffset += 4;
  dataView.setUint32(byteOffset, 0x004e4942, true);
  byteOffset += 4;

  for (const value of binaryBuffer) {
    dataView.setUint8(byteOffset, value);
    byteOffset += 1;
  }

  return glb;
}
